import os
import copy
import regis.rex_json
import regis.util
import regis.task_raii_printing
import regis.diagnostics
import requests
import zipfile 
import shutil
import threading
import argparse

from pathlib import Path

root = regis.util.find_root()
settings = regis.rex_json.load_file(os.path.join(root, "build", "config", "settings.json"))
build_dir = os.path.join(root, settings["build_folder"])
temp_dir = os.path.join(root, settings["intermediate_folder"])
tools_install_dir = os.path.join(temp_dir, settings["tools_folder"])
tool_paths_filepath = os.path.join(tools_install_dir, "tool_paths.json")
zip_downloads_path = os.path.join(tools_install_dir, "zips")

tool_paths_dict = {}
if os.path.exists(tool_paths_filepath):
  tool_paths_dict = regis.rex_json.load_file(tool_paths_filepath)
required_tools = []
not_found_tools = []

def __load_required_tools_dict():
  tools_required = []
  json_blob = regis.rex_json.load_file(os.path.join(root, "build", "config", "required_tools.json"))
  for object in json_blob:
    tools_required.append(json_blob[object])

  return tools_required

def __print_tool_found(tool, path : str):
  regis.diagnostics.log_no_color(f"{tool['config_name']} found at {path}")

def __get_tool_extension(tool):
  extension = "" 
  if "extension" in tool:
    extension = tool["extension"]
  elif regis.util.is_windows():
    extension = ".exe"

  return extension

def __look_for_tools(required_tools):
  paths = []
    
  for required_tool in required_tools:
    stem = required_tool["stem"]
    path = os.path.join(tools_install_dir, required_tool["archive_name"])
    version = regis.util.load_version_file(path)

    # check if our version is up to date
    if version != required_tool["version"]:
      regis.diagnostics.log_err(f"{stem} is out of date")
      not_found_tools.append(required_tool)
      continue

    config_name = required_tool["config_name"]

    # check if the tool path is already in the cached paths
    if config_name in tool_paths_dict:
      tool_path = tool_paths_dict[config_name]
      if (os.path.exists(tool_path)):
          regis.diagnostics.log_no_color(f"{stem} found at {tool_path}")
          continue
      else:
        regis.diagnostics.log_err(f"Error: tool path cached, but path doesn't exist: {tool_path}")
        not_found_tools.append(required_tool)
        continue

    # if not, add the path of the tool directory where it'd be downloaded to
    paths_to_use = copy.deepcopy(paths)
    paths_to_use.append(os.path.join(tools_install_dir, required_tool["path"]))

    # look for the tool
    exe_extension = __get_tool_extension(required_tool)
    tool_path = regis.util.find_file_in_paths(f"{stem}{exe_extension}", paths_to_use)

    # tool is found, add it to the cached paths
    if tool_path != '':
      __print_tool_found(required_tool, tool_path)
      tool_config_name = required_tool["config_name"]
      tool_paths_dict[tool_config_name] = tool_path
    # tool is not found, add it to the list to be looked for later
    else:
      not_found_tools.append(required_tool)

  return not_found_tools

def are_installed():
  task_print = regis.task_raii_printing.TaskRaiiPrint("Checking if tools are installed")

  global required_tools
  required_tools = __load_required_tools_dict()
  
  global tool_paths_dict
  if tool_paths_dict == None:
    tool_paths_dict = {}
    
  global not_found_tools
  not_found_tools = __look_for_tools(required_tools)

  if len(not_found_tools) == 0:
    regis.diagnostics.log_info("All tools found")
    regis.rex_json.save_file(tool_paths_filepath, tool_paths_dict)
    return True
  else:
    regis.diagnostics.log_warn(f"Tools that weren't found: ")
    for tool in not_found_tools:
      regis.diagnostics.log_warn(f"\t-{tool['stem']}")

  return False

def __download_file(url):
  filename = os.path.basename(url)
  filePath = os.path.join(zip_downloads_path, filename)
  
  if not os.path.exists(filePath):
    response = requests.get(url)
    open(filePath, "wb").write(response.content)
  
def __make_zip_download_path():
  if not os.path.exists(zip_downloads_path):
    os.makedirs(zip_downloads_path)

def __download_tool(name, version, numZipFiles):
  task_print = regis.task_raii_printing.TaskRaiiPrint(f"Downloading tool {name} {version}")

  threads = []
  for i in range(numZipFiles):
    threads.append(__launch_download_thread((f"https://github.com/RisingLiberty/RegisZip/raw/{version}/data/{name}.zip.{(i + 1):03d}")))

  for thread in threads:
    thread.join()

def __download_tools_archive():
  task_print = regis.task_raii_printing.TaskRaiiPrint("Downloading tools")

  # filter duplicate tools
  tools_to_download = []
  for not_found_tool in not_found_tools:
    archive_name = not_found_tool["archive_name"]
    should_add = True
    for tool_to_download in tools_to_download:
      if archive_name == tool_to_download["archive_name"]:
        should_add = False
        break
    
    if should_add:
      tools_to_download.append(not_found_tool)

  for not_found_tool in tools_to_download:
    arch_name = not_found_tool["archive_name"]
    num_zip_files = not_found_tool["num_zip_files"]
    version = not_found_tool["version"]
    __download_tool(arch_name, version, num_zip_files)

  return tools_to_download
    
def __enumerate_tools(zipsFolder):
  zips = os.listdir(zipsFolder)
  tools = []
  for zip in zips:
    stem = Path(zip).stem
    if stem not in tools:
      tools.append(stem)

  return tools

def __zip_files_for_tool(stem, folder):
  zips = os.listdir(folder)
  tool_zip_files = []
  for zip in zips:
    if Path(zip).stem == stem:
      tool_zip_files.append(os.path.join(folder, zip))

  return tool_zip_files

def __unzip_tools():
  task_print = regis.task_raii_printing.TaskRaiiPrint("Unzipping files")
  tools_to_unzip = __enumerate_tools(zip_downloads_path)

  for tool in tools_to_unzip:
    tool_zip_files = __zip_files_for_tool(tool, zip_downloads_path)
    tool_master_zip = os.path.join(zip_downloads_path, f"{tool}")
    with open(tool_master_zip, "ab") as f:
      for tool_zip in tool_zip_files:
        with open(tool_zip, "rb") as z:
            f.write(z.read())

    with zipfile.ZipFile(tool_master_zip, "r") as zip_obj:
        zip_obj.extractall(tools_install_dir)

  regis.diagnostics.log_info(f"tools unzipped to {tools_install_dir}")

def __create_version_files(toolsToDownload : []):
  for tool in toolsToDownload:
    path = os.path.join(tools_install_dir, tool["archive_name"])
    regis.util.create_version_file(path, tool["version"])

def __delete_tmp_folders():
  shutil.rmtree(zip_downloads_path)

def __launch_download_thread(url):
    thread = threading.Thread(target=__download_file, args=(url,))
    thread.start()
    return thread  

def download():
  task_print = regis.task_raii_printing.TaskRaiiPrint("Downloading tools")
  __make_zip_download_path()
  tools_to_download = __download_tools_archive()
  __unzip_tools()
  __create_version_files(tools_to_download)
  __delete_tmp_folders()

def install():
  task_print = regis.task_raii_printing.TaskRaiiPrint("installing tools")

  global tool_paths_dict
  global not_found_tools
  for tool in not_found_tools:

    # look for tool in the folder where it'd be downloaded to
    exe_extension = __get_tool_extension(tool)
    path = regis.util.find_file_in_folder(f"{tool['stem']}{exe_extension}", os.path.join(tools_install_dir, tool["path"]))

    # if not found, something is wrong and we have to investigate manually
    if path == "":
      tool_name = tool["stem"]
      regis.diagnostics.log_err(f"failed to find {tool_name}")
    else:
      # if found, add it to the cached paths
      __print_tool_found(tool, path)
      tool_config_name = tool["config_name"]
      tool_paths_dict[tool_config_name] = path
  
  # save cached paths to disk
  regis.rex_json.save_file(tool_paths_filepath, tool_paths_dict)

def run():
  if not are_installed():
    download()
    install()

if __name__ == "__main__":
  parser = argparse.ArgumentParser()
  args, unknown = parser.parse_known_args()

  run()




import os
import subprocess
import regis.required_tools
import regis.rex_json
import regis.util
import regis.diagnostics
import regis.subproc

from pathlib import Path

tool_paths_dict = regis.required_tools.tool_paths_dict

def __find_project_file(project):
  root = regis.util.find_root()
  project_file_name = f"{project}.nproj"
  settings = regis.rex_json.load_file(os.path.join(root, "build", "config", "settings.json"))
  intermediate_folder = settings["intermediate_folder"]
  build_folder = settings["build_folder"]

  directory = os.path.join(root, intermediate_folder, build_folder, "ninja")
  
  for root_, dirs, files in os.walk(directory):
    for file in files:
      if Path(file).name.lower() == project_file_name.lower():
        return os.path.join(root_, file)

  return ""

def __launch_new_build(project : str, config : str, compiler : str, shouldClean : bool, alreadyBuild : [str]):
  project_file_path = __find_project_file(project)

  if project_file_path == "":
    regis.diagnostics.log_err(f"project '{project}' was not found, have you generated it?")
    return 1, alreadyBuild
    
  json_blob = regis.rex_json.load_file(project_file_path)

  project_lower = project.lower()
  compiler_lower = compiler.lower()
  config_lower = config.lower()
  
  if compiler not in json_blob[project_lower]:
    regis.diagnostics.log_err(f"no compiler '{compiler}' found for project '{project}'")
    return 1, alreadyBuild
  
  if config not in json_blob[project_lower][compiler_lower]:
    regis.diagnostics.log_err(f"no config '{config}' found in project '{project}' for compiler '{compiler}'")
    return 1, alreadyBuild

  ninja_file = json_blob[project_lower][compiler_lower][config_lower]["ninja_file"]
  dependencies = json_blob[project_lower][compiler_lower][config_lower]["dependencies"]

  regis.diagnostics.log_info(f"Building: {project}")

  ninja_path = tool_paths_dict["ninja_path"]
  if shouldClean:
    proc = regis.subproc.run(f"{ninja_path} -f {ninja_file} -t clean")
    proc.wait()

  proc = regis.subproc.run(f"{ninja_path} -f {ninja_file}")
  proc.wait()
  return proc.returncode, alreadyBuild

def new_build(project : str, config : str, compiler : str, shouldClean : bool):
  already_build = []
  res, build_projects = __launch_new_build(project, config, compiler, shouldClean, already_build)
  return res
  


# ============================================ 
#
# Author: Nick De Breuck
# Twitter: @nick_debreuck
# 
# File: run_clang_tools.py
# Copyright (c) Nick De Breuck 2023
#
# ============================================

import os
import argparse
import regis.diagnostics
import regis.subproc
import regis.util
import regis.required_tools
import regis.rex_json
import shutil

clang_tidy_first_pass_filename = ".clang-tidy_first_pass"
clang_tidy_second_pass_filename = ".clang-tidy_second_pass"
clang_tidy_format_filename = ".clang-format"
root = regis.util.find_root()
settings = regis.rex_json.load_file(os.path.join(root, "build", "config", "settings.json"))
intermediate_folder = settings["intermediate_folder"]
build_folder = settings["build_folder"]
processes_in_flight_filename = os.path.join(root, intermediate_folder, build_folder, "ninja", "post_builds_in_flight.tmp")
project = ""

def __quoted_path(path):
  quote = "\""
  return f"{quote}{path}{quote}"

def __run_command(command):
  proc = regis.subproc.run(command)
  streamdata = proc.communicate()[0]
  return proc.returncode

def run(projectName, compdb, srcRoot):
  script_path = os.path.dirname(__file__)
  global project
  project = projectName

  headerFilters = regis.util.retrieve_header_filters(compdb, project)
  headerFiltersRegex = regis.util.create_header_filter_regex(headerFilters)

  clang_tidy_path = regis.required_tools.tool_paths_dict["clang_tidy_path"]
  clang_format_path = regis.required_tools.tool_paths_dict["clang_format_path"]
  clang_apply_replacements_path = regis.required_tools.tool_paths_dict["clang_apply_replacements_path"]
  clang_config_file = os.path.join(compdb, clang_tidy_first_pass_filename)

  regis.diagnostics.log_info("Running clang-tidy - auto fixes")
  rc = __run_command(f"py {__quoted_path(script_path)}/run_clang_tidy.py -clang-tidy-binary={__quoted_path(clang_tidy_path)} -clang-apply-replacements-binary={__quoted_path(clang_apply_replacements_path)} -config-file={__quoted_path(clang_config_file)} -p={__quoted_path(compdb)} -header-filter={headerFiltersRegex} -quiet -fix") # force clang compiler, as clang-tools expect it

  if rc != 0:
    raise Exception("clang-tidy auto fixes failed")

  regis.diagnostics.log_info("Running clang-format")
  rc = __run_command(f"py {__quoted_path(script_path)}/run_clang_format.py --clang-format-executable={__quoted_path(clang_format_path)} -r -i {srcRoot}")

  if rc != 0:
    raise Exception("clang-format failed")

if __name__ == "__main__":
  # arguments setups
  parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter)

  parser.add_argument("-p", "--project", help="project name")
  parser.add_argument("-compdb", help="compiler database folder")
  parser.add_argument("-srcroot", help="src root folder")
  
  args, unknown = parser.parse_known_args()

 # useful for debugging
  regis.diagnostics.log_info(f"Executing {__file__}")

 # execute the script
  run(args.project, args.compdb, args.srcroot)

 # print. We're done.
  regis.diagnostics.log_info("Done.")

  exit(0)


# ============================================ 
#
# Author: Nick De Breuck
# Twitter: @nick_debreuck
# 
# File: required_libs.py
# Copyright (c) Nick De Breuck 2022
#
# ============================================

# This script is specifically designed to all required libraries for Rex.

import os
import regis.task_raii_printing
import regis.util
import regis.rex_json
import regis.diagnostics
import threading
import requests
import zipfile 
import shutil
from pathlib import Path

root = regis.util.find_root()
settings = regis.rex_json.load_file(os.path.join(root, "build", "config", "settings.json"))
build_dir = os.path.join(root, settings["build_folder"])
temp_dir = os.path.join(root, settings["intermediate_folder"])
tools_install_dir = os.path.join(temp_dir, settings["tools_folder"])
libs_install_dir = os.path.join(temp_dir, settings["libs_folder"])
lib_paths_filepath = os.path.join(libs_install_dir, "lib_paths.json")
zip_downloads_path = os.path.join(libs_install_dir, "zips")

lib_paths_dict = {}
if os.path.exists(lib_paths_filepath):
  lib_paths_dict = regis.rex_json.load_file(lib_paths_filepath)
required_libs = []
not_found_libs = []

def __load_required_libs_dict():
  libs_required = []
  json_blob = regis.rex_json.load_file(os.path.join(root, "build", "config", "required_libs.json"))
  for object in json_blob:
    libs_required.append(json_blob[object])

  return libs_required

def __print_lib_found(lib_path, path : str):
  regis.diagnostics.log_no_color(f"{lib_path} found at {path}")

# finds any of the paths in the required lib and checks if they're cached already
# if they're not it adds them to a local list and returns that list
def __find_uncached_paths(lib):
  config_name = lib["config_name"]
  required_lib_paths = lib["paths"]
  cached_lib_paths = []
  if config_name in lib_paths_dict:
    cached_lib_paths = lib_paths_dict[config_name]
  
  lib_paths_to_search = []
  for lib_path in required_lib_paths:
    # first let's check if the path is already in the cached paths
    # if it's not in there, then we have to look for it later
    abs_path = regis.util.find_directory_in_paths(lib_path, cached_lib_paths)

    if abs_path == None:
      lib_paths_to_search.append(lib_path)
      continue

    # if it is there, check if exists, if not, we'll have to look for it later as well
    if not os.path.exists(abs_path):
      regis.diagnostics.log_warn(f"lib path cached but doesn't exist: {lib_path}")
      lib_paths_to_search.append(lib_path)
      continue

    # otherwise print that we've found the path
    __print_lib_found(lib_path, abs_path)
    
  return lib_paths_to_search

def __look_for_paths(lib, pathsToSearch : list[str], whereToSearch : list[str]):
  not_found_paths = []
  for path in pathsToSearch:
    abs_path = regis.util.find_directory_in_paths(path, whereToSearch)
    if abs_path == None:
      not_found_paths.append(path)
      continue

    __print_lib_found(path, abs_path)
    config_name = lib["config_name"]
    if config_name not in lib_paths_dict:
      lib_paths_dict[config_name] = [] 
    lib_paths_dict[config_name].append(abs_path)

  return not_found_paths

def __download_file(url):
  filename = os.path.basename(url)
  filepath = os.path.join(zip_downloads_path, filename)
  
  if not os.path.exists(filepath):
    response = requests.get(url)
    open(filepath, "wb").write(response.content)

def __launch_download_thread(url):
    thread = threading.Thread(target=__download_file, args=(url,))
    thread.start()
    return thread  

def __download_lib(name, version, numZipFiles):
  task_print = regis.task_raii_printing.TaskRaiiPrint(f"Downloading lib {name} {version}")

  threads = []
  for i in range(numZipFiles):
    threads.append(__launch_download_thread((f"https://github.com/RisingLiberty/RegisZip/raw/{version}/data/{name}.zip.{(i + 1):03d}")))

  for thread in threads:
    thread.join()

def __enumerate_libs(zipsFolder):
  zips = os.listdir(zipsFolder)
  libs = []
  for zip in zips:
    stem = Path(zip).stem
    if stem not in libs:
      libs.append(stem)

  return libs

def __enumerate_zip_files_for_lib(stem, folder):
  zips = os.listdir(folder)
  lib_zip_files = []
  for zip in zips:
    if Path(zip).stem == stem:
      lib_zip_files.append(os.path.join(folder, zip))

  return lib_zip_files

def __unzip_lib(name):
  task_print = regis.task_raii_printing.TaskRaiiPrint("Unzipping files")
  libs_to_unzip = __enumerate_libs(zip_downloads_path)

  for lib in libs_to_unzip:
    lib_zip_files = __enumerate_zip_files_for_lib(lib, zip_downloads_path)
    lib_master_zip = os.path.join(zip_downloads_path, f"{lib}.zip")
    with open(lib_master_zip, "ab") as f:
      for lib_zip in lib_zip_files:
        with open(lib_zip, "rb") as z:
            f.write(z.read())

    with zipfile.ZipFile(lib_master_zip, "r") as zip_obj:
        zip_obj.extractall(libs_install_dir)

  regis.diagnostics.log_info(f"libs unzipped to {libs_install_dir}")

def __is_up_to_date(installPaths, lib):
  for install_path in installPaths:
    path = os.path.join(install_path, lib["archive_name"])
    version = regis.util.load_version_file(path)
    if version == lib["version"]:
      return True
  
  return False
  
def __look_for_required_libs(required_libs):
  not_found_libs = []
  install_paths = regis.util.env_paths()
  install_paths.append(tools_install_dir)
  install_paths.append(libs_install_dir)
  for required_lib in required_libs:
    if not __is_up_to_date(install_paths, required_lib):
      regis.diagnostics.log_err(f"{required_lib['archive_name']} out of date")
      not_found_libs.append(required_lib)
      continue
    
    uncached_paths = __find_uncached_paths(required_lib)
    paths_not_found = __look_for_paths(required_lib, uncached_paths, install_paths)
    
    if len(paths_not_found) > 0:
      regis.diagnostics.log_warn("Couldn't find some paths")
      
      for path in paths_not_found:
        regis.diagnostics.log_warn(path)
      
      not_found_libs.append(required_lib)
            
  return not_found_libs

# checks all paths of the required libs, making sure all of them are installed
# if they're not installed, it'll flag a required_lib as not fully installed
def __are_installed():
  task_print = regis.task_raii_printing.TaskRaiiPrint("Checking if libs are installed")

  global required_libs
  required_libs = __load_required_libs_dict()
  
  global lib_paths_dict
  if lib_paths_dict == None:
    lib_paths_dict = {}
    
  global not_found_libs
  not_found_libs = __look_for_required_libs(required_libs)
  
  return len(not_found_libs) == 0

def __download():
  # create the temporary path for zips
  if not os.path.exists(zip_downloads_path):
      os.makedirs(zip_downloads_path)

  # filter duplicate tools
  libs_to_download = []
  for not_found_tool in not_found_libs:
    archive_name = not_found_tool["archive_name"]
    should_add = True
    for tool_to_download in libs_to_download:
      if archive_name == tool_to_download["archive_name"]:
        should_add = False
        break
    
    if should_add:
      libs_to_download.append(not_found_tool)

  for lib in libs_to_download:
    __download_lib(lib["archive_name"], lib["version"], lib["num_zip_files"])
    __unzip_lib(lib)
    regis.util.create_version_file(os.path.join(libs_install_dir, lib["archive_name"]), lib["version"])

  # remove it after all libs have been downloaded
  shutil.rmtree(zip_downloads_path)
  
def __install():
  for lib in not_found_libs:
    config_name = lib["config_name"]
    if config_name in lib_paths_dict:
      lib_paths_dict[config_name].clear()
    paths_not_found = __look_for_paths(lib, lib["paths"], [libs_install_dir])
  
    if len(paths_not_found) > 0:
      regis.diagnostics.log_err(f"failed to install {config_name}")
  
def run():
  if not __are_installed():
    __download()
    __install()

  regis.rex_json.save_file(lib_paths_filepath, lib_paths_dict)
  
if __name__ == "__main__":
  run()


# ============================================ 
#
# Author: Nick De Breuck
# Twitter: @nick_debreuck
# 
# File: test.py
# Copyright (c) Nick De Breuck 2023
#
# ============================================

import argparse
import os
import traceback
import shutil
import time
import regis.required_tools
import regis.util
import regis.task_raii_printing
import regis.rex_json
import regis.code_coverage
import regis.diagnostics
import regis.generation
import regis.build

from pathlib import Path
from datetime import datetime

root_path = regis.util.find_root()
tool_paths_dict = regis.required_tools.tool_paths_dict
settings = regis.rex_json.load_file(os.path.join(root_path, "build", "config", "settings.json"))
__pass_results = {}

def get_pass_results():
  return __pass_results

def __is_in_line(line : str, keywords : list[str]):
  for keyword in keywords:
    if keyword.lower() in line.lower():
      return True

  return False

def __default_output_callback(output):
  error_keywords = ["failed", "error"]
  warn_keywords = ["warning"]

  for line in iter(output.readline, b''):
    new_line : str = line.decode('UTF-8')
    if new_line.endswith('\n'):
      new_line = new_line.removesuffix('\n')

    if __is_in_line(new_line, error_keywords):
      regis.diagnostics.log_err(new_line)
      continue
    elif __is_in_line(new_line, warn_keywords):
      regis.diagnostics.log_warn(new_line)
      continue
    
    regis.diagnostics.log_no_color(new_line)

def __run_include_what_you_use(fixIncludes = False):
  task_print = regis.task_raii_printing.TaskRaiiPrint("running include-what-you-use")

  regis.generation.new_generation(os.path.join(root_path, "build", "config", "settings.json"), "")

  intermediate_folder = os.path.join(root_path, settings["intermediate_folder"], settings["build_folder"])
  result = regis.util.find_all_files_in_folder(intermediate_folder, "compile_commands.json")
    
  for compiler_db in result:
    iwyu_path = tool_paths_dict["include_what_you_use_path"]
    iwyu_tool_path = os.path.join(Path(iwyu_path).parent, "iwyu_tool.py")
    fix_includes_path = os.path.join(Path(iwyu_path).parent, "fix_includes.py")
    compiler_db_folder = Path(compiler_db).parent
    output_path = os.path.join(compiler_db_folder, "iwyu_output.log")
    os.system(f"py {iwyu_tool_path} -v -p={compiler_db_folder} > {output_path}") # needs to use os.system or iwyu will parse the command incorrectly
    if fixIncludes:
      os.system(f"py {fix_includes_path} --update_comments --safe_headers < {output_path}")
    
    regis.diagnostics.log_info(f"include what you use info saved to {output_path}")

  return 0

# the compdbPath directory contains all the files needed to configure clang tools
# this includes the compiler databse, clang tidy config files, clang format config files
# and a custom generated project file (in run_clang_tools.py), which should have the same filename as the source root directory
# of the project you're testing
def __get_project_name(compdbPath):
  dirs = os.listdir(compdbPath)
  for dir in dirs:
    if ".project" in dir:
      return dir.split(".")[0]
  
  return ""

def __run_clang_tidy():
  task_print = regis.task_raii_printing.TaskRaiiPrint("running clang-tidy")

  regis.generation.new_generation(os.path.join(root_path, "build", "config", "settings.json"), "")

  intermediate_folder = os.path.join(root_path, settings["intermediate_folder"], settings["build_folder"])
  result = regis.util.find_all_files_in_folder(intermediate_folder, "compile_commands.json")

  rc = 0
  for compiler_db in result:
    script_path = os.path.dirname(__file__)
    clang_tidy_path = tool_paths_dict["clang_tidy_path"]
    clang_apply_replacements_path = tool_paths_dict["clang_apply_replacements_path"]
    compiler_db_folder = Path(compiler_db).parent
    config_file_path = f"{compiler_db_folder}/.clang-tidy_second_pass"

    project_name = __get_project_name(compiler_db_folder)
    header_filters = regis.util.retrieve_header_filters(compiler_db_folder, project_name)
    header_filters_regex = regis.util.create_header_filter_regex(header_filters)

    cmd = f"py {script_path}/run_clang_tidy.py"
    cmd += f" -clang-tidy-binary={clang_tidy_path}"
    cmd += f" -clang-apply-replacements-binary={clang_apply_replacements_path}"
    cmd += f" -config-file={config_file_path}"
    cmd += f" -p={compiler_db_folder}"
    cmd += f" -header-filter={header_filters_regex}" # only care about headers of the current project
    cmd += f" -quiet"

    proc = regis.util.run_subprocess_with_callback(cmd, __default_output_callback)
    new_rc = regis.util.wait_for_process(proc)
    if new_rc != 0:
      regis.diagnostics.log_err(f"clang-tidy failed for {compiler_db}")
      regis.diagnostics.log_err(f"config file: {config_file_path}")
    rc |= new_rc

  return rc

def __generate_test_files(sharpmakeArgs):
  root = regis.util.find_root()
  settings_path = os.path.join(root, "build", "config", "settings.json")
  proc = regis.generation.new_generation(settings_path, sharpmakeArgs)
  proc.wait()
  return proc.returncode

def __find_projects_with_suffix(directory, suffix):
  projects = []
  for root, dirs, files in os.walk(directory):
    for file in files:
      filename = Path(file).name
      if filename.lower().endswith(f"{suffix}.nproj".lower()):
        projects.append(Path(filename).stem)

  return projects

def __build_test_files(projectSuffix : str, configs : [str], compilers : [str]):
  should_clean = False

  result = 0

  intermediate_folder = settings["intermediate_folder"]
  build_folder = settings["build_folder"]

  directory = os.path.join(root_path, intermediate_folder, build_folder, "ninja")
  projects = __find_projects_with_suffix(directory, projectSuffix)

  for project in projects:
    for config in configs:
      for compiler in compilers:
        result |= regis.build.new_build(project, config, compiler, should_clean)

  return result

def __find_test_programs(folder, regex):
  intermediate_folder = os.path.join(folder)
  regis.diagnostics.log_info(f"looking for executables in {os.path.join(root_path, intermediate_folder)}")
  result = regis.util.find_all_files_in_folder(os.path.join(root_path, intermediate_folder), regex)
  coverage_programs = []
  for res in result:
    if regis.util.is_executable(res):
      coverage_programs.append(res.absolute())

  return coverage_programs

# unit tests
def __generate_tests():
  task_print = regis.task_raii_printing.TaskRaiiPrint("generating unit test projects")
  return __generate_test_files("/generateUnitTests")

def __build_tests():
  task_print = regis.task_raii_printing.TaskRaiiPrint("building tests")
  return __build_test_files("test", ["debug", "debug_opt", "release"], ["msvc", "clang"])

def __run_unit_tests():
  task_print = regis.task_raii_printing.TaskRaiiPrint("running unit tests")
  unit_test_programs = __find_test_programs(os.path.join(settings["intermediate_folder"], settings["build_folder"], "ninja"), "*test*")
  
  rc = 0
  for program in unit_test_programs:
    regis.diagnostics.log_info(f"running: {Path(program).name}")
    proc = regis.util.run_subprocess(program)
    new_rc = regis.util.wait_for_process(proc)
    if new_rc != 0:
      regis.diagnostics.log_err(f"unit test failed for {program}") # use full path to avoid ambiguity
    rc |= new_rc

  return rc

# coverage
def __generate_coverage():
  task_print = regis.task_raii_printing.TaskRaiiPrint("generating coverage code")
  return __generate_test_files("/generateUnitTests /enableCoverage")

def __build_coverage():
  task_print = regis.task_raii_printing.TaskRaiiPrint("building coverage code")
  return __build_test_files("_coverage", ["coverage"], ["clang"])

def __run_coverage():
  task_print = regis.task_raii_printing.TaskRaiiPrint("running coverage")
  unit_test_programs = __find_test_programs(os.path.join(settings["intermediate_folder"], settings["build_folder"], "ninja"), "*coverage*")

  rc = 0
  for program in unit_test_programs:
    regis.diagnostics.log_info(f"running: {Path(program).name}")
    os.environ["LLVM_PROFILE_FILE"] = __get_coverage_rawdata_filename(program) # this is what llvm uses to set the raw data filename for the coverage data
    proc = regis.util.run_subprocess(program)
    new_rc = regis.util.wait_for_process(proc)
    if new_rc != 0:
      regis.diagnostics.log_err(f"unit test failed for {program}") # use full path to avoid ambiguity
    rc |= new_rc

  return unit_test_programs

def __relocate_coverage_data(programsRun : list[str]):
  task_print = regis.task_raii_printing.TaskRaiiPrint("relocating coverage files")
  data_files = []

  for program in programsRun:
    coverage_rawdata_filename = __get_coverage_rawdata_filename(program)
    newPath = os.path.join(Path(program).parent, coverage_rawdata_filename)
    if (os.path.exists(newPath)):
      os.remove(newPath)
    os.rename(coverage_rawdata_filename, newPath)
    data_files.append(newPath)
    
  return data_files

def __index_rawdata_files(rawdataFiles : list[str]):
  task_print = regis.task_raii_printing.TaskRaiiPrint("indexing rawdata files")
  output_files = []

  for file in rawdataFiles:
    output_files.append(regis.code_coverage.create_index_rawdata(file))

  return output_files

def __create_coverage_reports(programsRun, indexdataFiles):
  task_print = regis.task_raii_printing.TaskRaiiPrint("creating coverage reports")

  rc = 0
  for index in range(len(programsRun)):
    program = programsRun[index]
    indexdata_file = indexdataFiles[index]

    if Path(program).stem != Path(indexdata_file).stem:
      rc = 1
      regis.diagnostics.log_err(f"program stem doesn't match coverage file stem: {Path(program).stem} != {Path(indexdata_file).stem}")

    regis.code_coverage.create_line_oriented_report(program, indexdata_file)
    regis.code_coverage.create_file_level_summary(program, indexdata_file)
    regis.code_coverage.create_lcov_report(program, indexdata_file)

  return rc

def __parse_coverage_reports(indexdataFiles):
  task_print = regis.task_raii_printing.TaskRaiiPrint("parsing coverage reports")

  rc = 0
  for indexdata_file in indexdataFiles:
    report_filename = regis.code_coverage.get_file_level_summary_filename(indexdata_file)
    rc |= regis.code_coverage.parse_file_summary(report_filename)

  return rc

def __get_coverage_rawdata_filename(program : str):
  return f"{Path(program).stem}.profraw"

# asan
def __generate_address_sanitizer():
  task_print = regis.task_raii_printing.TaskRaiiPrint("generating address sanitizer code")
  return __generate_test_files("/generateUnitTests /enableAddressSanitizer")

def __build_address_sanitizer():
  task_print = regis.task_raii_printing.TaskRaiiPrint("building address sanitizer code")
  return __build_test_files("_asan", ["address_sanitizer"], ["clang"])

def __run_address_sanitizer():
  task_print = regis.task_raii_printing.TaskRaiiPrint("running address sanitizer tests")
  unit_test_programs = __find_test_programs(os.path.join(settings["intermediate_folder"], settings["build_folder"], "ninja"), "*asan*")
  
  rc = 0
  for program in unit_test_programs:
    regis.diagnostics.log_info(f"running: {Path(program).name}")
    log_folder_path = Path(program).parent
    log_folder = log_folder_path.as_posix()
    
    # for some reason, setting an absolute path for the log folder doesn't work
    # so we have to set the working directory of the program to where it's located so the log file will be there as well
    # ASAN_OPTIONS common flags: https://github.com/google/sanitizers/wiki/SanitizerCommonFlags
    # ASAN_OPTIONS flags: https://github.com/google/sanitizers/wiki/AddressSanitizerFlags
    asan_options = f"print_stacktrace=1:log_path=asan.log"
    os.environ["ASAN_OPTIONS"] = asan_options # print callstacks and save to log file
    
    proc = regis.util.run_subprocess_with_working_dir(program, log_folder)
    new_rc = regis.util.wait_for_process(proc)
    log_file_path = os.path.join(log_folder, f"asan.log.{proc.pid}")
    if new_rc != 0 or os.path.exists(log_file_path):
      regis.diagnostics.log_err(f"address sanitizer failed for {program}") # use full path to avoid ambiguity
      regis.diagnostics.log_err(f"for more info, please check: {log_file_path}")
      new_rc = 1
    rc |= new_rc

  return rc

# ubsan
def __generate_undefined_behavior_sanitizer():
  task_print = regis.task_raii_printing.TaskRaiiPrint("generating undefined behavior sanitizer code")
  return __generate_test_files("/generateUnitTests /enableUBSanitizer")

def __build_undefined_behavior_sanitizer():
  task_print = regis.task_raii_printing.TaskRaiiPrint("building undefined behavior sanitizer code")
  return __build_test_files("_ubsan", ["undefined_behavior_sanitizer"], ["clang"])

def __run_undefined_behavior_sanitizer():
  task_print = regis.task_raii_printing.TaskRaiiPrint("running undefined behavior sanitizer tests")
  unit_test_programs = __find_test_programs(os.path.join(settings["intermediate_folder"], settings["build_folder"], "ninja"), "*ubsan*")
  
  rc = 0
  for program in unit_test_programs:
    regis.diagnostics.log_info(f"running: {Path(program).name}")
    log_folder_path = Path(program).parent
    log_folder = log_folder_path.as_posix()
    
    # for some reason, setting an absolute path for the log folder doesn't work
    # so we have to set the working directory of the program to where it's located so the log file will be there as well
    # UBSAN_OPTIONS common flags: https://github.com/google/sanitizers/wiki/SanitizerCommonFlags
    ubsan_options = f"print_stacktrace=1:log_path=ubsan.log"
    os.environ["UBSAN_OPTIONS"] = ubsan_options # print callstacks and save to log file
    proc = regis.util.run_subprocess_with_working_dir(program, log_folder)
    new_rc = regis.util.wait_for_process(proc)
    log_file_path = os.path.join(log_folder, f"ubsan.log.{proc.pid}")
    if new_rc != 0 or os.path.exists(log_file_path): # if there's a ubsan.log.pid created, the tool found issues
      regis.diagnostics.log_err(f"undefined behavior sanitizer failed for {program}") # use full path to avoid ambiguity
      regis.diagnostics.log_err(f"for more info, please check: {log_file_path}")
      new_rc = 1
    rc |= new_rc

  return rc

# fuzzy
def __generate_fuzzy_testing():
  task_print = regis.task_raii_printing.TaskRaiiPrint("generating fuzzy testing code")
  return __generate_test_files("/enableFuzzyTesting")

def __build_fuzzy_testing():
  task_print = regis.task_raii_printing.TaskRaiiPrint("building fuzzy testing code")
  return __build_test_files("_fuzzy", ["fuzzy"], ["clang"])

def __run_fuzzy_testing():
  task_print = regis.task_raii_printing.TaskRaiiPrint("running fuzzy tests")
  fuzzy_programs = __find_test_programs(os.path.join(settings["intermediate_folder"], settings["build_folder"], "ninja"), "*fuzzy*")
  
  rc = 0
  for program in fuzzy_programs:
    regis.diagnostics.log_info(f"running: {Path(program).name}")
    log_folder_path = Path(program).parent
    log_folder = log_folder_path.as_posix()
    
    # for some reason, setting an absolute path for the log folder doesn't work
    # so we have to set the working directory of the program to where it's located so the log file will be there as well
    # Can't use both ASAN as well as UBSAN options, so we'll set the same for both and hope that works
    # https://gcc.gnu.org/bugzilla/show_bug.cgi?id=94328
    # https://stackoverflow.com/questions/60774638/logging-control-for-address-sanitizer-plus-undefined-behavior-sanitizer
    asan_options = f"print_stacktrace=1:log_path=fuzzy.log"
    ubsan_options = f"print_stacktrace=1:log_path=fuzzy.log"
    os.environ["ASAN_OPTIONS"] = asan_options # print callstacks and save to log file
    os.environ["UBSAN_OPTIONS"] = ubsan_options # print callstacks and save to log file
    num_runs = 10000 # we'll run 10'000 fuzzy tests, should be more than enough
    proc = regis.util.run_subprocess_with_working_dir(f"{program} -runs={num_runs}", log_folder)
    new_rc = regis.util.wait_for_process(proc)
    log_file_path = os.path.join(log_folder, f"fuzzy.log.{proc.pid}")
    if new_rc != 0 or os.path.exists(log_file_path): # if there's a ubsan.log.pid created, the tool found issues
      regis.diagnostics.log_err(f"fuzzy testing failed for {program}") # use full path to avoid ambiguity
      if os.path.exists(log_file_path):
        regis.diagnostics.log_err(f"issues found while fuzzing!")
        regis.diagnostics.log_err(f"for more info, please check: {log_file_path}")
      new_rc = 1
    rc |= new_rc

  return rc

def test_include_what_you_use():
  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  rc = __run_include_what_you_use()

  if rc != 0:
    regis.diagnostics.log_err(f"include-what-you-use pass failed")

  __pass_results["include-what-you-use"] = rc

def test_clang_tidy():
  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  rc = __run_clang_tidy() # works
  if rc != 0:
    regis.diagnostics.log_err(f"clang-tidy pass failed")

  __pass_results["clang-tidy"] = rc

def test_unit_tests():
  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  rc = __generate_tests() # works
  if rc != 0:
    regis.diagnostics.log_err(f"failed to generate tests")
  __pass_results["unit tests generation"] = rc

  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  rc |= __build_tests() # works
  if rc != 0:
    regis.diagnostics.log_err(f"failed to build tests")
  __pass_results["unit tests building"] = rc

  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  rc |= __run_unit_tests() # works
  if rc != 0:
    regis.diagnostics.log_err(f"unit tests failed")
  __pass_results["unit tests result"] = rc

def test_code_coverage():
  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  rc = __generate_coverage() # works
  if rc != 0:
    regis.diagnostics.log_err(f"failed to generate coverage")
  __pass_results["coverage generation"] = rc

  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  rc = __build_coverage() # works
  if rc != 0:
    regis.diagnostics.log_err(f"failed to build coverage")
  __pass_results["coverage building"] = rc

  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  programs_run = __run_coverage() # works
  
  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  rawdata_files = __relocate_coverage_data(programs_run) # works
  
  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  indexdata_files = __index_rawdata_files(rawdata_files) # works
  
  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  rc |= __create_coverage_reports(programs_run, indexdata_files) # works
  if rc != 0:
    regis.diagnostics.log_err(f"failed to create coverage reports")
  __pass_results["coverage report creation"] = rc

  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  rc |= __parse_coverage_reports(indexdata_files) # works
  if rc != 0:
    regis.diagnostics.log_err(f"Not all the code was covered")
  __pass_results["coverage report result"] = rc

def test_asan():
  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  rc = __generate_address_sanitizer() # works
  if rc != 0:
    regis.diagnostics.log_err(f"failed to generate asan code")
  __pass_results["address sanitizer generation"] = rc

  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  rc |= __build_address_sanitizer() # works
  if rc != 0:
    regis.diagnostics.log_err(f"failed to build asan code")
  __pass_results["address sanitizer building"] = rc
  
  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  rc |= __run_address_sanitizer() # works
  if rc != 0:
    regis.diagnostics.log_err(f"invalid code found with asan")
  __pass_results["address sanitizer result"] = rc

def test_ubsan():
  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  rc = __generate_undefined_behavior_sanitizer() # works
  if rc != 0:
    regis.diagnostics.log_err(f"failed to generate ubsan code")
  __pass_results["undefined behavior sanitizer generation"] = rc
  
  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  rc |= __build_undefined_behavior_sanitizer() # works
  if rc != 0:
    regis.diagnostics.log_err(f"failed to build ubsan code")
  __pass_results["undefined behavior sanitizer building"] = rc
  
  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  rc |= __run_undefined_behavior_sanitizer() # works
  if rc != 0:
    regis.diagnostics.log_err(f"invalid code found with ubsan")
  __pass_results["undefined behavior sanitizer result"] = rc

def test_fuzzy_testing():
  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  rc = __generate_fuzzy_testing() # works
  if rc != 0:
    regis.diagnostics.log_err(f"failed to generate fuzzy code")
  __pass_results["fuzzy testing generation"] = rc
  
  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  rc |= __build_fuzzy_testing() # works
  if rc != 0:
    regis.diagnostics.log_err(f"failed to build fuzzy code")
  __pass_results["fuzzy testing building"] = rc
  
  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  rc |= __run_fuzzy_testing() # works
  if rc != 0:
    regis.diagnostics.log_err(f"invalid code found with fuzzy")
  __pass_results["fuzzy testing result"] = rc

def __shutil_error(func, path, exec_info):
  regis.diagnostics.log_err(f"shutil error: {func}, {path}, {exec_info}")

def clean():
  intermediate_tests_path = os.path.join(root_path, settings["intermediate_folder"], settings["build_folder"])
  if os.path.exists(intermediate_tests_path):
    shutil.rmtree(intermediate_tests_path, onerror=__shutil_error)



import os
import argparse
import regis.util
import regis.rex_json
import regis.required_tools
import regis.required_libs
import shutil

root_path = regis.util.find_root()
settings = regis.rex_json.load_file(os.path.join(root_path, "build", "config", "settings.json"))
intermediate_dir = os.path.join(regis.util.find_root(), settings["intermediate_folder"])

def __clean_intermediate():
  # this clean the entire intermediate directory and all sub folders
  if os.path.exists(intermediate_dir):
    shutil.rmtree(intermediate_dir)

def run(shouldClean):
  if shouldClean:
    __clean_intermediate()
      
  regis.required_tools.run(False)
  regis.required_libs.run()
      
if __name__ == "__main__":
  parser = argparse.ArgumentParser()
  parser.add_argument("-clean", help="clean setup, as if run for the first time", action="store_true")
  args, unknown = parser.parse_known_args()

  run(args.clean)


import os
import regis.diagnostics
import regis.rex_json
import regis.util
import regis.required_tools
import regis.subproc
import regis.diagnostics

from pathlib import Path

root = regis.util.find_root()
settings = regis.rex_json.load_file(os.path.join(root, "build", "config", "settings.json"))
temp_dir = os.path.join(root, settings["intermediate_folder"])
tools_install_dir = os.path.join(temp_dir, settings["tools_folder"])
tool_paths_filepath = os.path.join(tools_install_dir, "tool_paths.json")
tool_paths_dict = regis.rex_json.load_file(tool_paths_filepath)

def __find_sharpmake_files(directory):
  sharpmakes_files = []
  for root, dirs, files in os.walk(directory):
    for file in files:
      extensions = Path(file).suffixes
      if len(extensions) == 2:
        if extensions[0] == ".sharpmake" and extensions[1] == ".cs":
          path = os.path.join(root, file)
          sharpmakes_files.append(path)
  
  return sharpmakes_files

def __find_sharpmake_root_files(directory):
  sharpmakes_files = []
  for root, dirs, files in os.walk(directory):
    for file in files:
      extensions = Path(file).suffixes
      if len(extensions) == 1:
        if extensions[0] == ".cs":
          path = os.path.join(root, file)
          sharpmakes_files.append(path)

  return sharpmakes_files

def __load_sharpmake_files(sharpmakePath : str):
  settings = regis.rex_json.load_file(sharpmakePath)
  sharpmake_root = os.path.join(root, settings["build_folder"], "sharpmake")
  source_root = os.path.join(root, settings["source_folder"])
  tests_root = os.path.join(root, settings["tests_folder"])
  
  sharpmakes_files = []
  sharpmakes_files.extend(__find_sharpmake_root_files(sharpmake_root))
  sharpmakes_files.extend(__find_sharpmake_files(source_root))
  sharpmakes_files.extend(__find_sharpmake_files(tests_root))

  return sharpmakes_files

def new_generation(settingsPath : str, sharpmakeArgs : list[str]):
  sharpmake_files = __load_sharpmake_files(settingsPath)
  
  sharpmake_path = tool_paths_dict["sharpmake_path"]
  if len(sharpmake_path) == 0:
    regis.diagnostics.log_err("Failed to find sharpmake path")
    return

  sharpmake_sources = ""
  for sharpmake_file in sharpmake_files:
    sharpmake_sources += "\""
    sharpmake_sources += sharpmake_file
    sharpmake_sources += "\", "

  sharpmake_sources = sharpmake_sources[0:len(sharpmake_sources) - 2]
  sharpmake_sources = sharpmake_sources.replace('\\', '/')

  return regis.subproc.run(f"{sharpmake_path} /sources({sharpmake_sources}) /diagnostics {sharpmakeArgs}")



import re
import subprocess
import regis.diagnostics

def __has_word(line, word : str):
  regex_str = f"({word.lower()})[\]}} \)\:]|({word.upper()})[\]}} \)\:]"
  result = re.findall(regex_str, line)
  return len(result) > 0

def __has_error(line : str):
  return __has_word(line, "error") or __has_word(line, "errors") or __has_word(line, "failed")

def __has_warning(line):
  return __has_word(line, "warning") or __has_word(line, "warnings")

def __build_output_callback(output : bytes):
  for line in iter(output.readline, b''):
    new_line : str = line.decode('UTF-8')
    if new_line.endswith('\n'):
      new_line = new_line.removesuffix('\n')

    if __has_error(new_line):
      regis.diagnostics.log_err(new_line)
    elif __has_warning(new_line):
      regis.diagnostics.log_warn(new_line)
    else:
      regis.diagnostics.log_no_color(new_line)

def run(cmd):
  proc = subprocess.Popen(cmd, stdout=subprocess.PIPE)
  __build_output_callback(proc.stdout)
  return proc



import os
import json
import requests
import zipfile
import shutil
from enum import Enum

import regis.diagnostics
import regis.util
import regis.rex_json

class Host(Enum):
    UNKNOWN = 0
    GITLAB = 1
    GITHUB = 2

def __get_script_path():
    return os.path.dirname(os.path.realpath(__file__))

def __get_host(path):
    if "gitlab" in path:
        return Host.GITLAB
    elif "github" in path:
        return Host.GITHUB
    
    regis.diagnostics.log_info("Unknown host!")
    return Host.UNKNOWN

def __build_gitlab_path(baseUrl, name, tag):
    url = os.path.join(baseUrl, "-")
    url = os.path.join(url, "archive")
    url = os.path.join(url, tag)
    url = os.path.join(url, name+"-"+tag+".zip")
    url = url.replace("\\", "/")
    return url
def __build_github_path(baseUrl, tag):
    url = os.path.join(baseUrl, "archive")
    url = os.path.join(url, "refs")
    url = os.path.join(url, "tags")
    url = os.path.join(url, tag+".zip")
    url = url.replace("\\", "/")
    return url
def __build_host_path(baseUrl, name, tag):
    host = __get_host(baseUrl)
    if host == Host.GITHUB:
        return __build_github_path(baseUrl, tag)
    elif host == Host.GITLAB:
        return __build_gitlab_path(baseUrl, name, tag)
    else:
        regis.diagnostics.log_err(f"Unknown url host: {host} in url: {baseUrl}")
        return ""

def __load_externals_required():
    root = regis.util.find_root()

    json_blob = regis.rex_json.load_file(os.path.join(root, "build", "config", "required_externals.json"))
    if json_blob == None:
        regis.diagnostics.log_err("Loaded json blob is None, stopping json parse")
        return []

    externals_required = []
    for object in json_blob:
        externals_required.append(json_blob[object])

    return externals_required

def __download_external(url):
    # get basename of the URL (a.k.a. the filename + extention we would like to download)
    url_basename = os.path.basename(url)

    # request a download of the given URL
    if not os.path.exists(url_basename):
        response = requests.get(url)
        if response.status_code == requests.codes.ok:
            # write the downloaded file to disk
            open(url_basename, "wb").write(response.content)
        else:
            # bad request was made
            regis.diagnostics.log_err(f"Bad request [{str(response.status_code)}] for given url: {url}")
            return []
        
    # extract the zip file on disk
    # we cache the files within the directory before 
    # and after extraction, this gives us the ability
    # to examine the added files within the directory
    regis.diagnostics.log_info("Extracting: " + url)
    
    # pre list directories
    # cached directories before we downloaded anything
    pre_list_dir = os.listdir(__get_script_path())
    with zipfile.ZipFile(url_basename,"r") as zip_ref:
        zip_ref.extractall(__get_script_path())

    # post list directories
    # directories after we downloaded the repository
    post_list_dir = os.listdir(__get_script_path())

    regis.diagnostics.log_info("Looking for added extracted directories ...")
    added_directory_names = []
    for post_dir in post_list_dir:
        count = pre_list_dir.count(post_dir)
        if count == 0:
            added_directory_names.append(post_dir)
    regis.diagnostics.log_info(f"Found ({str(len(added_directory_names))}): ".join(added_directory_names))

    # remove the created zip file
    os.remove(url_basename)

    return added_directory_names

def __verify_external(externalPath, requiredTag):
    external_name = os.path.basename(externalPath)

    if os.path.exists(externalPath):
        regis.diagnostics.log_no_color(f"External found: {external_name}")
        regis.diagnostics.log_no_color(f"validating version ...")
        version = regis.util.load_version_file(externalPath)
        if version != requiredTag:
            regis.diagnostics.log_err(f"Invalid version data found, redownloading external: {external_name}")
            return False
        return True

    else:
        return False

def __install_external(external):
    root = regis.util.find_root()

    external_url = external["url"]
    external_name = external["name"]
    external_tag = external["tag"]
    external_store = external["storage"]
    external_store = external_store.replace("~", root)

    externals_dir = os.path.join(external_store, external_name)

    # if the external is already present we need to check if we need to redownload anything
    valid_external = __verify_external(externals_dir, external_tag)
    if not valid_external:    
        # any data that was already available will be deleted 
        # the data will be out of date anyway when a download is triggered
        if os.path.exists(externals_dir):
            shutil.rmtree(externals_dir)

        url = __build_host_path(external_url, external_name, external_tag)
        added_directories = __download_external(url)     

        if len(added_directories) == 1:
            # move to output directory
            shutil.move(os.path.join(__get_script_path(), added_directories[0]), os.path.join(external_store, added_directories[0]))
            # change directory name
            cwd = os.getcwd()
            os.chdir(external_store)
            os.rename(added_directories[0], external_name)
            os.chdir(cwd)
        elif len(added_directories) > 1:
            # create output directory
            if not os.path.exists(externals_dir):
                os.makedirs(externals_dir)
            # move to output directory
            for added_directory in added_directories:
                shutil.move(os.path.join(__get_script_path(), added_directory), externals_dir)
        else:
            regis.diagnostics.log_err("No directories where extracted.")
            return

        regis.util.create_version_file(externals_dir, external_tag)   

def run():
    regis.diagnostics.log_info("Start installing externals ...")

    externals_required = __load_externals_required()
    if externals_required == None:
        regis.diagnostics.log_err("Required externals is None, exiting ...")
        return

    for external in externals_required:
        __install_external(external)    
